(* archive access *)
(* solyga@gmx.de, 2024-07-08 *)
IMPLEMENTATION MODULE Pak;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT IO, FIO, FIOStd, Str, Lib, Lib2, Lib3, SO, Config, Pak0;

TYPE
  TElmDesc = Pak0.TElmDesc;

VAR
  pak: FIO.File;
  wrt: BOOLEAN;  (* writing allowed *)
  hdr: Pak0.THdrDesc;
  toc: ARRAY TVldIdx OF TElmDesc;
  pos: ARRAY TVldIdx OF TLen;
  cnt: TCnt;
  top: TLen;
  mod: BOOLEAN;  (* toc modified *)
  map: RECORD
    ofs: ARRAY TVldIdx OF TVldIdx;
    ref: ARRAY TVldIdx OF TVldIdx;
    len: ARRAY TVldIdx OF TVldIdx;
  END;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  (*
  IO.WrStr( Config.data.prg );
  IO.WrStr( ': ' );
  *)
  IO.WrStr( 'Pak.' );
  IO.WrStr( s );
  IO.WrLn();
END Message;

PROCEDURE Warning( r-: ARRAY OF CHAR );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'WARNING Pak.' );
  SO.WrStr( b, r );
  IO.WrStr( s );
  IO.WrLn();
END Warning;

PROCEDURE Error( r-: ARRAY OF CHAR );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'ERROR Pak.' );
  SO.WrStr( b, r );
  IO.WrStr( s );
  IO.WrLn();
END Error;

PROCEDURE Fatal( r-: ARRAY OF CHAR );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'FATAL Pak.' );
  SO.WrStr( b, r );
  Lib3.FatalError( s );
  HALT;
END Fatal;


PROCEDURE DoSetPos( p: TLen );
BEGIN
  FIO.Seek( pak, p );
  IF FIO.GetPos( pak ) # p THEN Fatal( 'DoSetPos: Seek failed' ) END;
END DoSetPos;

PROCEDURE DoRead( VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;
VAR n: CARDINAL;
BEGIN
  n := FIO.RdBin( pak, x, l );
  RETURN n;
END DoRead;

PROCEDURE DoWrite( x-: ARRAY OF BYTE; l: CARDINAL );
BEGIN
  IF ~ wrt THEN Fatal( 'DoWrite: Archive is read-only' ) END;
  FIO.WrBin( pak, x, l );
END DoWrite;


PROCEDURE Cnt(): TCnt;
BEGIN
  RETURN cnt;
END Cnt;

PROCEDURE GetRef( i: TVldIdx; VAR r: TRef );
BEGIN
  IF i > cnt THEN Lib2.Fill( r, 0 ); RETURN END;
  r := toc[i].ref;
END GetRef;

PROCEDURE Ofs( i: TVldIdx ): TLen;
BEGIN
  IF i > cnt THEN (* Fatal( 'Len: Invalid index' ) *) RETURN 0 END;
  RETURN toc[i].rng.ofs;
END Ofs;

PROCEDURE Len( i: TVldIdx ): TLen;
BEGIN
  IF i > cnt THEN (* Fatal( 'Len: Invalid index' ) *) RETURN 0 END;
  RETURN toc[i].rng.len;
END Len;


PROCEDURE Find( r-: TRef ): TIdx;
VAR i: TIdx;
BEGIN
  FOR i := 1 TO cnt DO
    IF toc[i].ref = r THEN RETURN i END;
  END;
  RETURN 0;
END Find;

PROCEDURE Add( r-: TRef; l: TLen ): TIdx;
VAR i: TIdx;
BEGIN
  i := Find( r );
  IF i > 0 THEN Fatal( 'Add: Reference already exists' ) END;
  IF cnt >= cntMax THEN Fatal( 'Add: Too many elements' ) END;
  i := cnt + 1;
  toc[i].ref     := r;
  toc[i].rng.ofs := top;  (* top up-to-date? !!! *)
  toc[i].rng.len := l;
  cnt := i;
  mod := TRUE;
  IF l = 0 THEN Warning( 'Add: Zero file length' ) END;
  RETURN i;
END Add;

PROCEDURE Read( i: TVldIdx; VAR x: ARRAY OF BYTE; n: CARDINAL ): CARDINAL;
VAR l, p: TLen;
BEGIN
  IF i > cnt THEN RETURN 0 END;
  IF n > SIZE( x ) THEN n := SIZE( x ) END;
  l := toc[i].rng.len;
  p := pos[i];
  IF p > l THEN p := l END;
  IF TLen( n ) > l - p THEN n := CARDINAL( l - p ) END;
  (* ( p <= l ) & ( n <= l - p ) *)
  DoSetPos( toc[i].rng.ofs + p );
  n := DoRead( x, n );
  pos[i] := p + TLen( n );
  RETURN n;
END Read;

PROCEDURE Write( i: TVldIdx; x-: ARRAY OF BYTE; n: CARDINAL ): CARDINAL;
VAR l, p: TLen;
BEGIN
  IF ~ wrt THEN Fatal( 'Write: Archive is read-only' ) END;
  IF i > cnt THEN RETURN 0 END;
  IF n > SIZE( x ) THEN n := SIZE( x ) END;
  l := toc[i].rng.len;
  p := pos[i];
  IF p > l THEN p := l END;
  IF TLen( n ) > l - p THEN n := CARDINAL( l - p ) END;
  (* ( p <= l ) & ( n <= l - p ) *)
  DoSetPos( toc[i].rng.ofs + p );
  DoWrite( x, n );
  pos[i] := p + TLen( n );
  RETURN n;
END Write;


PROCEDURE Pos( i: TVldIdx ): TLen;
BEGIN
  IF i > cnt THEN RETURN 0 END;
  RETURN pos[i];
END Pos;

PROCEDURE SetPos( i: TVldIdx; p: TLen );
VAR l: TLen;
BEGIN
  IF i > cnt THEN RETURN END;
  l := toc[i].rng.len;
  IF p > l THEN p := l END;
  pos[i] := p;
END SetPos;


PROCEDURE Check();

  PROCEDURE Contains( x-, y-: Pak0.TRngDesc ): BOOLEAN;
  BEGIN
    (* x contains y *)
    RETURN ( x.ofs <= y.ofs ) & ( x.len >= y.len ) & ( ( y.ofs - x.ofs ) <= ( x.len - y.len ) );
  END Contains;

  PROCEDURE Overlap( x-, y-: Pak0.TRngDesc ): BOOLEAN;
  BEGIN
    (* x and y overlap *)
    RETURN ( x.ofs <= y.ofs ) & ( y.ofs - x.ofs < x.len  ) OR
           ( y.ofs <= x.ofs ) & ( x.ofs - y.ofs < y.len  );
  END Overlap;

  PROCEDURE GetRng( i: TIdx; VAR x: Pak0.TRngDesc );
  BEGIN
    IF i = 0 THEN
      x := hdr.toc;
    ELSIF i <= cnt THEN
      x := toc[i].rng;
    ELSE
      Lib2.Fill( x, 0 );
    END;
  END GetRng;

  PROCEDURE WrRng( VAR b: SO.TBuffer; r-: Pak0.TRngDesc );
  VAR c: CHAR;
  BEGIN
    c := b.c.p;
    b.c.p := ' ';
    SO.WrLngHex( b, r.ofs, 8 );
    SO.WrChar( b, ' ' );
    SO.WrLngHex( b, r.len, 8 );
    b.c.p := c;
  END WrRng;

VAR
  e: BOOLEAN;
  d, ri, rj: Pak0.TRngDesc;
  i, j: TIdx;

  PROCEDURE ErrRng();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    e := TRUE;
    SO.Init( b, s );
    SO.WrStr( b, 'Check: Element ' );
    SO.WrCard( b, i, 1 );
    SO.WrStr( b, ' out of data range: ' );
    WrRng( b, d );
    SO.WrStr( b, ' ' );
    WrRng( b, ri );
    Error( s );
  END ErrRng;

  PROCEDURE ErrOvl();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    e := TRUE;
    SO.Init( b, s );
    SO.WrStr( b, 'Check: Element ' );
    SO.WrCard( b, i, 1 );
    SO.WrStr( b, ' overlaps element ' );
    SO.WrCard( b, j, 1 );
    SO.WrStr( b, ': ' );
    WrRng( b, ri );
    SO.WrStr( b, ' ' );
    WrRng( b, rj );
    Error( s );
  END ErrOvl;

  PROCEDURE Show();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Check: ' );
    IF e THEN
      SO.WrStr( b, 'Failed' );
    ELSE
      SO.WrStr( b, 'Passed' );
    END;
    Message( s );
  END Show;

BEGIN
  e := FALSE;
  d.ofs := SIZE( hdr );
  d.len := FIO.Size( pak ) - SIZE( hdr );
  FOR i := 0 TO cnt DO
    GetRng( i, ri );
    IF ~ Contains( d, ri ) THEN ErrRng() END;
    FOR j := i + 1 TO cnt DO
      GetRng( j, rj );
      IF Overlap( ri, rj ) THEN ErrOvl() END;
    END;
  END;
  Show();
END Check;


PROCEDURE RefIdx( is: TVldIdx ): TVldIdx;
BEGIN
  IF is > cnt THEN Fatal( 'RefIdx: Invalid index' ) END;
  RETURN map.ref[ is ];
END RefIdx;

PROCEDURE OfsIdx( is: TVldIdx ): TVldIdx;
BEGIN
  IF is > cnt THEN Fatal( 'OfsIdx: Invalid index' ) END;
  RETURN map.ofs[ is ];
END OfsIdx;

PROCEDURE LenIdx( is: TVldIdx ): TVldIdx;
BEGIN
  IF is > cnt THEN Fatal( 'LenIdx: Invalid index' ) END;
  RETURN map.len[ is ];
END LenIdx;

PROCEDURE FindRef( x-: TRef; VAR is: TIdx ): TIdx;
VAR
  a, b, m: CARDINAL;
BEGIN
  a := 1;
  b := 1 + cnt;
  WHILE a < b DO
    m := a + ( b - a ) DIV 2;  (* m <= cnt *)
    IF Str.Compare( toc[ RefIdx(m) ].ref, x ) <= 0 THEN a := m + 1 ELSE b := m END;
  END;  (* a = b; x[a-1] <= x < x[a] *)
  (* uncomment if TIdx should become sub-range type
  IF a > MAX( TIdx ) THEN is := 0; RETURN 0 END;
  *)
  is := a;
  IF a <= cnt THEN
    a := RefIdx( a );
    IF Str.Compare( toc[a].ref, x ) = 0 THEN RETURN a END;
  END;
  RETURN 0;
END FindRef;

PROCEDURE FindOfs( x: TLen; VAR is: TIdx ): TIdx;
VAR
  a, b, m: CARDINAL;
BEGIN
  a := 1;
  b := 1 + cnt;
  WHILE a < b DO
    m := a + ( b - a ) DIV 2;  (* m <= cnt *)
    IF toc[ OfsIdx(m) ].rng.ofs <= x THEN a := m + 1 ELSE b := m END;
  END;  (* a = b; x[a-1] <= x < x[a] *)
  (* uncomment if TIdx should become sub-range type
  IF a > MAX( TIdx ) THEN is := 0; RETURN 0 END;
  *)
  is := a;
  IF a <= cnt THEN
    a := OfsIdx( a );
    IF toc[a].rng.ofs = x THEN RETURN a END;
  END;
  RETURN 0;
END FindOfs;

PROCEDURE FindLen( x: TLen; VAR is: TIdx ): TIdx;
VAR
  a, b, m: CARDINAL;
BEGIN
  a := 1;
  b := 1 + cnt;
  WHILE a < b DO
    m := a + ( b - a ) DIV 2;  (* m <= cnt *)
    IF toc[ LenIdx(m) ].rng.len <= x THEN a := m + 1 ELSE b := m END;
  END;  (* a = b; x[a-1] <= x < x[a] *)
  (* uncomment if TIdx should become sub-range type
  IF a > MAX( TIdx ) THEN is := 0; RETURN 0 END;
  *)
  is := a;
  IF a <= cnt THEN
    a := LenIdx( a );
    IF toc[a].rng.len = x THEN RETURN a END;
  END;
  RETURN 0;
END FindLen;

PROCEDURE InsRefAt( is: TVldIdx );
VAR
  n: TCnt;
BEGIN
  n := cnt - is;
  IF n > 0 THEN Lib.Move( ADR(map.ref[is]), ADR(map.ref[is+1]), n*SIZE(TVldIdx) ) END;
  map.ref[ is ] := cnt;
END InsRefAt;

PROCEDURE InsOfsAt( is: TVldIdx );
VAR
  n: TCnt;
BEGIN
  n := cnt - is;
  IF n > 0 THEN Lib.Move( ADR(map.ofs[is]), ADR(map.ofs[is+1]), n*SIZE(TVldIdx) ) END;
  map.ofs[ is ] := cnt;
END InsOfsAt;

PROCEDURE InsLenAt( is: TVldIdx );
VAR
  n: TCnt;
BEGIN
  n := cnt - is;
  IF n > 0 THEN Lib.Move( ADR(map.len[is]), ADR(map.len[is+1]), n*SIZE(TVldIdx) ) END;
  map.len[ is ] := cnt;
END InsLenAt;

PROCEDURE Map();
VAR
  n: TCnt;
  e: TVldIdx;
  r, o, l: TIdx;
  i, j, k: TIdx;
BEGIN
  Lib2.Fill( map, 0 );
  n := cnt;
  cnt := 0;
  WHILE cnt < n DO
    e := cnt + 1;
    i := FindRef( toc[e].ref,     r );
    j := FindOfs( toc[e].rng.ofs, o );
    k := FindLen( toc[e].rng.len, l );
    INC( cnt );
    InsRefAt( r );
    InsOfsAt( o );
    InsLenAt( l );
  END;
END Map;

PROCEDURE Load( n-: ARRAY OF CHAR );
VAR
  len: TLen;

  PROCEDURE Chr( c: CHAR ): CHAR;
  BEGIN
    IF ( c < ' ' ) OR ( '~' < c ) THEN c := '.' END;
    RETURN c;
  END Chr;

  PROCEDURE Show1();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Load: "' );
    SO.WrStr( b, n );
    SO.WrStr( b, '" len ' );
    b.c.p := '0';
    SO.WrLngHex( b, len, 8 );
    Message( s );
  END Show1;

  PROCEDURE Show2();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
    i: CARDINAL;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Load: id "' );
    FOR i := 0 TO 3 DO
      SO.WrChar( b, Chr(hdr.id[i]) );
    END;
    SO.WrStr( b, '" ofs ' );
    b.c.p := '0';
    SO.WrLngHex( b, hdr.toc.ofs, 8 );
    SO.WrStr( b, ' len ' );
    SO.WrLngHex( b, hdr.toc.len, 8 );
    b.c.p := ' ';
    SO.WrStr( b, ' cnt ' );
    SO.WrLngCard( b, cnt, 1 );
    Message( s );
  END Show2;

BEGIN
  (* open *)
  IF ~ FIO.Exists( n ) THEN Fatal( 'Load: Archive does not exist' ) END;
  pak := FIO.Open( n );
  len := FIO.Size( pak );
  Show1();
  (* load header *)
  IF DoRead( hdr, SIZE( hdr ) ) # SIZE( hdr ) THEN Fatal( 'Load: Cannot read header' ) END;
  cnt := hdr.toc.len DIV SIZE( TElmDesc );
  Show2();
  IF hdr.id # Pak0.idPak                  THEN Fatal( 'Load: Invalid file ID' ) END;
  IF hdr.toc.ofs < SIZE( hdr )            THEN Fatal( 'Load: Invalid TOC offset (too small)' ) END;
  IF hdr.toc.ofs > len                    THEN Fatal( 'Load: Invalid TOC offset (too large)' ) END;
  IF hdr.toc.len > len - hdr.toc.ofs      THEN Fatal( 'Load: Invalid TOC length (too large for this archive)' ) END;
  IF hdr.toc.len > SIZE( toc )            THEN Fatal( 'Load: Invalid TOC length (too large for this program)' ) END;
  IF hdr.toc.len MOD SIZE( TElmDesc ) # 0 THEN Fatal( 'Load: Invalid TOC length (not multiple of 40H)' ) END;
  (* load toc *)
  DoSetPos( hdr.toc.ofs );
  IF DoRead( toc, hdr.toc.len ) # hdr.toc.len THEN Fatal( 'Load: Cannot read TOC' ) END;
  Map();
  (*
  top := toc[cnt].rng.ofs + toc[cnt].rng.len;
  *)
  Check();
END Load;

PROCEDURE Create( n-: ARRAY OF CHAR );
VAR
  len: TLen;

  PROCEDURE Show();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Create: "' );
    SO.WrStr( b, n );
    SO.WrStr( b, '" len ' );
    b.c.p := '0';
    SO.WrLngHex( b, len, 8 );
    b.c.p := ' ';
    SO.WrStr( b, ' cnt ' );
    SO.WrLngCard( b, cnt, 1 );
    Message( s );
  END Show;

BEGIN
  IF ~ wrt THEN Fatal( 'Create: Archive is read-only' ) END;
  pak := FIO.Create( n );
  hdr.id := Pak0.idPak;
  hdr.toc.ofs := SIZE( hdr );
  hdr.toc.len := 0;
  cnt := 0;
  top := SIZE( hdr );
  DoWrite( hdr, SIZE( hdr ) );
  len := SIZE( hdr );
  Show();
  Check();
END Create;


PROCEDURE Show();
CONST
  ena = FALSE;
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
  i, is: TVldIdx;
BEGIN
  IF ~ ena THEN RETURN END;
  IF cnt = 0 THEN RETURN END;
  FOR is := 1 TO cnt DO
    i := OfsIdx( is );
    SO.Init( b, s );
    SO.WrCard( b, is, 3 );
    SO.WrStr( b, ' ' );
    SO.WrCard( b, i, 3 );
    SO.WrStr( b, ' ' );
    SO.WrLngHex( b, toc[i].rng.ofs, 8 );
    SO.WrStr( b, ' ' );
    SO.WrLngHex( b, toc[i].rng.len, 8 );
    SO.WrStr( b, ' ' );
    SO.WrStr( b, toc[i].ref );
    IO.WrStr( s );
    IO.WrLn();
  END;
END Show;

VAR
  oldTerm: PROC;

PROCEDURE DoneModule();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF ~ Config.data.vrb THEN RETURN END;
  SO.Init( b, s );
  SO.WrStr( b, 'DoneModule' );
  Message( s );
  oldTerm();
END DoneModule;

BEGIN
  pak := FIOStd.inv;
  wrt := FALSE;
  Lib2.Fill( hdr, 0 );
  Lib2.Fill( toc, 0 );
  Lib2.Fill( pos, 0 );
  cnt := 0;
  top := 0;
  mod := FALSE;
  Lib2.Fill( map, 0 );

  CASE Config.data.act OF
  | Config.actCheck:
      Load( Config.data.arc );
      Check();
  | Config.actList :
      Load( Config.data.arc );
  | Config.actTable:
      Load( Config.data.arc );
(*
  | Config.actCreate:
      Create( Config.data.arc );
      wrt := TRUE;
*)
  ELSE
      Load( Config.data.arc );
      Check();
  END;
  Show();

  Lib3.Terminate( DoneModule, oldTerm );
END Pak.
